import matplotlib.pyplot as plt
import numpy as np


def sample(x, y, nb_per_class=5):
    sampled_x, sampled_y = [], []

    for class_id in np.unique(y):
        indexes = np.where(y == class_id)[0][:nb_per_class]
        sampled_x.append(x[indexes])
        sampled_y.append(y[indexes])
    sampled_x = np.concatenate(sampled_x)
    sampled_y = np.concatenate(sampled_y)

    return sampled_x, sampled_y


def plot(dataset, figsize=None, path=None, nb_per_class=5):
    x, y = sample(dataset.x, dataset.y, nb_per_class=nb_per_class)
    if not dataset.open_image and x.shape[1] == 1:
        x = x.squeeze(1)

    c = 1
    for class_id in range(dataset.nb_classes):
        for sample_id in range(nb_per_class):
            ax = plt.subplot(dataset.nb_classes, nb_per_class, c)
            if dataset.open_image:
                img = dataset.get_image(class_id * nb_per_class + sample_id)
                ax.imshow(np.asarray(img))
            elif len(x.shape) == 3:  # Grayscale, no channel dimension
                ax.imshow(x[class_id * nb_per_class + sample_id], cmap="gray")
            else:
                ax.imshow(x[class_id * nb_per_class + sample_id])

            ax.set_xticks([])
            ax.set_yticks([])
            c += 1

    plt.subplots_adjust(bottom=0.01, top=0.5, wspace=0.00, hspace=0.0001)
    if path is None:
        plt.show()
    else:
        plt.savefig(path, bbox_inches='tight', pad_inches=0)



import abc
from typing import List, Tuple, Union

import numpy as np

from clloader.datasets.base import BaseDataset, PyTorchDataset
from clloader.datasets.pytorch import (CIFAR10, CIFAR100, KMNIST, MNIST, FashionMNIST)
from torchvision import datasets as torchdata
from torchvision import transforms


class Fellowship(BaseDataset):

    def __init__(
        self, data_path: str = "", download: bool = True, dataset_list: List[BaseDataset] = None
    ):
        super().__init__(data_path, download)

        self.datasets = [dataset(data_path, download) for dataset in dataset_list]

    def init(self) -> Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
        x_train, y_train = [], []
        x_test, y_test = [], []
        class_counter = 0

        for dataset in self.datasets:
            train, test = dataset.init()

            x_train.append(train[0])
            x_test.append(test[0])

            y_train.append(train[1] + class_counter)
            y_test.append(test[1] + class_counter)

            class_counter += len(np.unique(train[1]))

        x_train = np.concatenate(x_train)
        x_test = np.concatenate(x_test)
        y_train = np.concatenate(y_train)
        y_test = np.concatenate(y_test)

        return (x_train, y_train), (x_test, y_test)


class MNISTFellowship(Fellowship):

    def __init__(self, data_path: str = "", download: bool = True) -> None:
        super().__init__(data_path, download, dataset_list=[MNIST, FashionMNIST, KMNIST])


class CIFARFellowship(Fellowship):

    def __init__(self, data_path: str = "", download: bool = True) -> None:
        super().__init__(data_path, download, dataset_list=[CIFAR10, CIFAR100])



from clloader.datasets.base import (BaseDataset, ImageFolderDataset,
                                    InMemoryDataset, PyTorchDataset)
from clloader.datasets.fellowship import (CIFARFellowship, Fellowship,
                                          MNISTFellowship)
from clloader.datasets.imagenet import ImageNet100, ImageNet1000
from clloader.datasets.pytorch import (CIFAR10, CIFAR100, EMNIST, KMNIST,
                                       MNIST, QMNIST, FashionMNIST)
from clloader.datasets.transformed import PermutedMNIST

# yapf: disable
__all__ = [
    BaseDataset,
    PyTorchDataset,
    InMemoryDataset,
    ImageFolderDataset,
    CIFAR10,
    CIFAR100,
    MNIST,
    FashionMNIST,
    KMNIST,
    EMNIST,
    QMNIST,
    ImageNet100,
    ImageNet1000,
    PermutedMNIST,
    Fellowship,
    MNISTFellowship,
    CIFARFellowship
]



from typing import List, Tuple, Union

import numpy as np

from clloader.datasets import MNIST
from torchvision import datasets as torchdata
from torchvision import transforms


class PermutedMNIST(MNIST):

    def __init__(self, *args, nb_permutations=4, **kwargs):
        super().__init__(*args, **kwargs)

        self.nb_permutations = nb_permutations

    def init(self):
        base_train, base_test = super().init()

        x_train, y_train = [base_train[0]], [base_train[1]]
        x_test, y_test = [base_test[0]], [base_test[1]]

        class_counter = np.max(base_train[1]) + 1
        nb_base_class = class_counter

        for i in range(self.nb_permutations):
            permuted_train, permuted_test = self._permut(base_train[0], base_test[0], i)

            x_train.append(permuted_train)
            x_test.append(permuted_test)

            y_train.append(base_train[1] + class_counter)
            y_test.append(base_test[1] + class_counter)

            class_counter += nb_base_class + 1

        x_train = np.concatenate(x_train)
        y_train = np.concatenate(y_train)
        x_test = np.concatenate(x_test)
        y_test = np.concatenate(y_test)

        return (x_train, y_train), (x_test, y_test)

    def _permut(self, x_train, x_test, i):
        random_state = np.random.RandomState(seed=i)
        permutations = random_state.permutation(x_train.shape[1] * x_train.shape[2])

        train_shape = x_train.shape
        test_shape = x_test.shape

        x_train = x_train.reshape((train_shape[0], -1))[..., permutations].reshape(train_shape)
        x_test = x_test.reshape((test_shape[0], -1))[..., permutations].reshape(test_shape)

        return x_train, x_test



import warnings
from typing import Callable, List, Tuple, Union

import numpy as np
from PIL import Image

import torch
from clloader.datasets import BaseDataset
from clloader.viz import plot
from torchvision import transforms


class Dataset(torch.utils.data.Dataset):

    def __init__(
        self, x: np.ndarray, y: np.ndarray, trsf: transforms.Compose, open_image: bool = False
    ):
        self.x, self.y = x, y
        self.trsf = trsf
        self.open_image = open_image

    @property
    def nb_classes(self):
        return len(np.unique(self.y))

    def add_memory(self, x_memory: np.ndarray, y_memory: np.ndarray):
        self.x = np.concatenate((self.x, x_memory))
        self.y = np.concatenate((self.y, y_memory))

    def plot(self, path=None, figsize=None, nb_per_class=5):
        plot(self, figsize=figsize, path=path, nb_per_class=nb_per_class)

    def __len__(self):
        return self.x.shape[0]

    def get_image(self, index):
        x = self.x[index]
        if self.open_image:
            img = Image.open(x).convert("RGB")
        else:
            img = Image.fromarray(x.astype("uint8"))
        return img

    def __getitem__(self, index):
        img = self.get_image(index)
        y = self.y[index]
        img = self.trsf(img)
        return img, y


class CLLoader:

    def __init__(
        self,
        cl_dataset: BaseDataset,
        increment: Union[List[int], int],
        initial_increment: int = 0,
        train_transformations: List[Callable] = None,
        common_transformations: List[Callable] = None,
        evaluate_on="seen"
    ) -> None:
        self.cl_dataset = cl_dataset

        if train_transformations is None:
            train_transformations = []
        if common_transformations is None:
            common_transformations = self.cl_dataset.transformations
        self.train_trsf = transforms.Compose(train_transformations + common_transformations)
        self.test_trsf = transforms.Compose(common_transformations)

        if evaluate_on not in ("seen", "current", "all"):
            raise NotImplementedError(f"Evaluate mode {evaluate_on} is not supported.")
        self.evaluate_on = evaluate_on

        self._setup()

        self.increments = self.define_increments(increment, initial_increment)

    def define_increments(self, increment: Union[List[int], int],
                          initial_increment: int) -> List[int]:
        if isinstance(increment, list):
            return increment
        increments = []
        if initial_increment:
            increments.append(initial_increment)

        nb_tasks = (self.nb_classes - initial_increment) / increment
        if not nb_tasks.is_integer():
            raise Exception(
                "The tasks won't have an equal number of classes"
                f" with {len(self.class_order)} and increment {increment}"
            )
        increments.extend([increment for _ in range(int(nb_tasks))])

        return increments

    def _setup(self) -> None:
        self.train_data, self.test_data = self.cl_dataset.init()
        unique_classes = np.unique(self.train_data[1])

        self.class_order = self.cl_dataset.class_order or list(range(len(unique_classes)))

    @property
    def nb_classes(self) -> int:
        return len(np.unique(self.train_data[1]))

    @property
    def nb_tasks(self) -> int:
        return len(self)

    def __len__(self) -> int:
        """Returns the number of tasks.

        :return: Number of tasks.
        """
        return len(self.increments)

    def __iter__(self):
        self._counter = 0
        return self

    def __next__(self) -> Tuple[Dataset, Dataset]:
        if self._counter >= len(self):
            raise StopIteration
        task = self[self._counter]
        self._counter += 1
        return task

    def __getitem__(self, task_index):
        max_class = sum(self.increments[:task_index + 1])
        min_class = sum(self.increments[:task_index])  # 0 when task_index == 0.

        train = self._select_data(min_class, max_class)
        train_dataset = Dataset(*train, self.train_trsf, open_image=not self.cl_dataset.in_memory)

        # TODO: validation
        if self.evaluate_on == "seen":
            test = self._select_data(0, max_class, split="test")
        elif self.evaluate_on == "current":
            test = self._select_data(min_class, max_class, split="test")
        else:  # all
            test = self._select_data(0, self.nb_classes, split="test")

        test_dataset = Dataset(*test, self.test_trsf, open_image=not self.cl_dataset.in_memory)

        return train_dataset, test_dataset

    def _select_data(self, min_class, max_class, split="train"):
        if split == "train":
            x, y = self.train_data
        elif split == "val":
            pass  # TODO
        else:
            x, y = self.test_data

        indexes = np.where(np.logical_and(y >= min_class, y < max_class))[0]
        return x[indexes], y[indexes]



import abc
import os
from typing import List, Tuple, Union

import numpy as np

from torchvision import datasets as torchdata
from torchvision import transforms


class BaseDataset(abc.ABC):

    def __init__(self, data_path: str = "", download: bool = True) -> None:
        self.data_path = data_path
        self.download = download

    @abc.abstractmethod
    def init(self) -> Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
        pass

    @property
    def class_order(self) -> List[int]:
        return None

    @property
    def in_memory(self):
        return True

    @property
    def transformations(self):
        return [transforms.ToTensor()]


class PyTorchDataset(BaseDataset):
    dataset_type = None

    def init(self) -> Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
        train_dataset = self.dataset_type(self.data_path, download=self.download, train=True)
        x_train, y_train = np.array(train_dataset.data), np.array(train_dataset.targets)
        test_dataset = self.dataset_type(self.data_path, download=self.download, train=False)
        x_test, y_test = np.array(test_dataset.data), np.array(test_dataset.targets)

        return (x_train, y_train), (x_test, y_test)


class InMemoryDataset(BaseDataset):

    def __init__(
        self, x_train: np.ndarray, y_train: np.ndarray, x_test: np.ndarray, y_test: np.ndarray,
        **kwargs
    ):
        super().__init__(**kwargs)

        self.train = (x_train, y_train)
        self.test = (x_test, y_test)

    def init(self) -> Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
        return self.train, self.test


class ImageFolderDataset(BaseDataset):

    def __init__(self, train_folder: str, test_folder: str, download: bool = True, **kwargs):
        super().__init__(download=download, **kwargs)

        self.train_folder = train_folder
        self.test_folder = test_folder

        if download:
            self._download()

    @property
    def in_memory(self):
        return False

    def _download(self):
        pass

    def init(self) -> Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
        train_dataset = torchdata.ImageFolder(self.train_folder)
        train_data = self._format(train_dataset.imgs)

        test_dataset = torchdata.ImageFolder(self.test_folder)
        test_data = self._format(test_dataset.imgs)

        return train_data, test_data

    @staticmethod
    def _format(raw_data: List[Tuple[str, int]]) -> Tuple[np.ndarray, np.ndarray]:
        x = np.empty(len(raw_data), dtype="S255")
        y = np.empty(len(raw_data), dtype=np.int16)

        for i, (path, target) in enumerate(raw_data):
            x[i] = path
            y[i] = target

        return x, y



import numpy as np
import pytest

from clloader import CLLoader
from clloader.datasets import InMemoryDataset
from torch.utils.data import DataLoader


def gen_data():
    x_train = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_train = []
    for i in range(10):
        y_train.append(np.ones(2) * i)
    y_train = np.concatenate(y_train)

    x_test = np.random.randint(0, 255, size=(20, 32, 32, 3))
    y_test = np.copy(y_train)

    return (x_train, y_train), (x_test, y_test)


# yapf: disable

@pytest.mark.parametrize("increment,initial_increment,nb_tasks", [
    (2, 0, 5),
    (5, 0, 2),
    (1, 5, 6),
    (2, 4, 4),
    ([5, 1, 1, 3], 0, 4)
])
def test_increments(increment, initial_increment, nb_tasks):
    train, test = gen_data()
    dummy = InMemoryDataset(*train, *test)
    clloader = CLLoader(dummy, increment, initial_increment)

    assert clloader.nb_tasks == nb_tasks
    seen_tasks = 0

    for task_id, (train_dataset, test_dataset) in enumerate(clloader):
        seen_tasks += 1

        if isinstance(increment, list):
            max_class = sum(increment[:task_id + 1])
            min_class = sum(increment[:task_id])
        elif initial_increment:
            max_class = initial_increment + increment * task_id
            min_class = initial_increment + increment * (task_id -1) if task_id > 0 else 0
        else:
            max_class = increment * (task_id + 1)
            min_class = increment * task_id

        for _ in DataLoader(train_dataset):
            pass
        for _ in DataLoader(test_dataset):
            pass

        assert np.max(train_dataset.y) == max_class - 1
        assert np.min(train_dataset.y) == min_class
        assert np.max(test_dataset.y) == max_class - 1
        assert np.min(test_dataset.y) == 0
    assert seen_tasks == nb_tasks



import abc
from typing import List, Tuple, Union

import numpy as np

from clloader.datasets import PyTorchDataset
from torchvision import datasets as torchdata
from torchvision import transforms


class CIFAR10(PyTorchDataset):
    dataset_type = torchdata.cifar.CIFAR10

    @property
    def transformations(self):
        return [
            transforms.ToTensor(),
            transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010))
        ]


class CIFAR100(PyTorchDataset):
    dataset_type = torchdata.cifar.CIFAR100

    @property
    def transformations(self):
        return [
            transforms.ToTensor(),
            transforms.Normalize((0.5071, 0.4867, 0.4408), (0.2675, 0.2565, 0.2761))
        ]


class MNIST(PyTorchDataset):
    dataset_type = torchdata.MNIST


class FashionMNIST(PyTorchDataset):
    dataset_type = torchdata.FashionMNIST


class KMNIST(PyTorchDataset):
    dataset_type = torchdata.KMNIST


class EMNIST(PyTorchDataset):
    dataset_type = torchdata.EMNIST


class QMNIST(PyTorchDataset):
    dataset_type = torchdata.QMNIST



import abc
import os
from typing import List, Tuple, Union

import numpy as np

from clloader.datasets import ImageFolderDataset
from torchvision import datasets as torchdata


class ImageNet1000(ImageFolderDataset):

    def _download(self):
        if not os.path.exists(self.train_folder) or not os.path.exists(self.test_folder):
            raise IOError(
                "You must download yourself the ImageNet dataset."
                " Please go to http://www.image-net.org/challenges/LSVRC/2012/downloads and"
                " download 'Training images (Task 1 & 2)' and 'Validation images (all tasks)'."
            )


class ImageNet100(ImageNet1000):

    def __init__(
        self, *args, train_subset: Union[Tuple[np.array, np.array], str],
        test_subset: Union[Tuple[np.array, np.array], str], **kwargs
    ):
        super().__init__(*args, **kwargs)

        self.train_subset = train_subset
        self.test_subset = test_subset

    def init(self) -> Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]]:
        train = self._parse_subset(self.train_subset, train=True)
        test = self._parse_subset(self.test_subset, train=False)

        return train, test

    def _parse_subset(self,
                      subset: Union[Tuple[np.array, np.array], str],
                      train: bool = True) -> Tuple[np.array, np.array]:
        if isinstance(subset, str):
            x, y = [], []
            folder = self.train_folder if train else self.test_folder

            with open(subset, "r") as f:
                for line in f:
                    split_line = line.split(" ")
                    path = "/".join(split_line[0].strip().split("/")[1:])
                    x.append(os.path.join(folder, path))
                    y.append(int(split_line[1].strip()))
            x = np.array(x)
            y = np.array(y)
            return x, y
        return subset



import os
import tempfile

import numpy as np
import pytest
from PIL import Image

from clloader import CLLoader
from clloader.datasets import ImageFolderDataset
from torch.utils.data import DataLoader


def gen_imagefolder(path, nb_classes=10, nb_samples=2):
    for class_id in range(nb_classes):
        folder = os.path.join(path, f"class_{class_id}")
        os.makedirs(folder)

        for sample_id in range(nb_samples):
            arr = np.random.randint(0, 255, size=(32, 32, 3))
            img = Image.fromarray(arr.astype("uint8"))
            img.save(os.path.join(folder, f"sample_{sample_id}.jpg"))


# yapf: disable

@pytest.mark.parametrize("increment,initial_increment,nb_tasks", [
    (2, 0, 5),
    (5, 0, 2),
    (1, 5, 6),
    (2, 4, 4),
    ([5, 1, 1, 3], 0, 4)
])
def test_increments(increment, initial_increment, nb_tasks):
    with tempfile.TemporaryDirectory() as train_path, tempfile.TemporaryDirectory() as test_path:
        gen_imagefolder(train_path)
        gen_imagefolder(test_path)

        clloader = CLLoader(ImageFolderDataset(train_path, test_path), increment, initial_increment)

        assert clloader.nb_tasks == nb_tasks
        seen_tasks = 0
        for task_id, (train_dataset, test_dataset) in enumerate(clloader):
            seen_tasks += 1

            if isinstance(increment, list):
                max_class = sum(increment[:task_id + 1])
                min_class = sum(increment[:task_id])
            elif initial_increment:
                max_class = initial_increment + increment * task_id
                min_class = initial_increment + increment * (task_id -1) if task_id > 0 else 0
            else:
                max_class = increment * (task_id + 1)
                min_class = increment * task_id

            for _ in DataLoader(train_dataset):
                pass
            for _ in DataLoader(test_dataset):
                pass

            assert np.max(train_dataset.y) == max_class - 1
            assert np.min(train_dataset.y) == min_class
            assert np.max(test_dataset.y) == max_class - 1
            assert np.min(test_dataset.y) == 0
    assert seen_tasks == nb_tasks
